/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Letter.h"

/**
 * \brief Initializes an empty Letter.
 */
Letter::Letter () :
    QTableWidgetItem ()
{
    setTextAlignment ( Qt::AlignCenter );
    setFlags ( Qt::ItemIsSelectable | Qt::ItemIsEnabled );

    comment_ = "";
    versionInfo_ = "";
    block_ = "";
}

/**
 * \brief Copy constructor.
 *
 * Creates a new object with the same attributes as the given one.
 *
 * \param other Letter object to copy.
 */
Letter::Letter ( const Letter* other )
{
    setTextAlignment ( Qt::AlignCenter );
    setFlags ( Qt::ItemIsSelectable | Qt::ItemIsEnabled );

    setCodepoint ( other->getCodepoint () );
    setComment ( other->getComment () );
    setCollations ( other->getCollations () );
    setVersionInfo ( other->getVersionInfo () );
    setBlock ( other->getBlock () );
}

/**
 * \brief Adds a collation.
 *
 * \param collation Collation.
 */
void Letter::addCollation ( Collation collation )
{
    collations_.append ( collation );
}

/**
 * \brief Compares Letter objects by their linestring.
 *
 * The comparison by linestring does include codepoint, collation and comment.
 * It ignores version and block.
 *
 * \param other The other letter.
 * \return True if both linestrings are equal.
 */

bool Letter::equals ( const Letter& other ) const
{
    bool result = false;
    if ( other.getLinestring () == getLinestring () )
    {
        result = true;
    }
    return result;
}

/**
 * \brief Gets the block this letter belongs to.
 *
 * \return The blocks name.
 */
QString Letter::getBlock () const
{
    return block_;
}

/**
 * \brief Gets the codepoint of the letter.
 *
 * \return QList<uint> Codepoint.
 */
QList<uint> Letter::getCodepoint () const
{
    return codepoint_;
}

/**
 * \brief Gets the codepoint as hex string.
 *
 * \return Codepoint as hex string.
 */
QString Letter::getCodepointAsHex () const
{
    QString codepointAsHex = "";
    QString tmp;
    for ( int i = 0; i < codepoint_.size (); i++ )
    {
        tmp = QString::number ( getCodepoint ().at ( i ), 16 );

        while ( tmp.length () < 4 )
        {
            tmp.prepend ( '0' );
        }

        codepointAsHex.append ( tmp );
        codepointAsHex.append ( " " );
    }
    return codepointAsHex.trimmed ();
}

/**
 * \brief Gets the codepoint in UTF-8 as string.
 *
 * \return Codepoint in UTF-8 as string.
 */
QString Letter::getCodepointAsUtf8 () const
{
    QString codepointUtf8 = "";
    QByteArray utf8;
    for ( int i = 0; i < codepoint_.size (); ++i )
    {
        if ( i > 0 )
        {
            codepointUtf8.append ( " " );
        }
        utf8 = UCTools::unicodeToString ( codepoint_.at ( i ) ).toUtf8 ();
        for ( int j = 0; j < utf8.size (); ++j )
        {
            if ( j > 0 )
            {
                codepointUtf8.append ( " " );
            }
            codepointUtf8.append (
                                   QString::number (
                                                     QChar ( utf8.at ( j ) ).unicode (),
                                                     16 ) );
        }
    }
    return codepointUtf8;
}

/**
 * \brief Gets the collations.
 *
 * \return QList<Collation> The collation of the letter.
 */
QList<Collation> Letter::getCollations () const
{
    return collations_;
}

/**
 * \brief Gets the comment.
 *
 * \return The comment of the Letter.
 */
QString Letter::getComment () const
{
    return comment_;
}

/**
 * \brief Gets the Linestring of this letter.
 *
 * \return Letter as a formatted string.
 */
QString Letter::getLinestring () const
{
    QString linestring = "";

    for ( int i = 0; i < codepoint_.size (); ++i )
    {
        if ( i > 0 )
        {
            linestring.append ( " " );
        }
        linestring.append ( UCTools::formatCodepoint ( codepoint_.at ( i ) ) );
    }

    if ( linestring.size () < 5 )
    {
        linestring.append ( " " );
    }
    linestring.append ( " ; " );

    for ( int i = 0; i < collations_.count (); i++ )
    {
        linestring.append ( collations_.at ( i ).toString () );
    }

    linestring.append ( " # " );
    linestring.append ( comment_ );
    return linestring;
}

/**
 * \brief Same as getLinestring(), but with the codepoint encoded in UTF-8.
 *
 * \return Letter as a formatted string.
 */
QString Letter::getLinestringUtf8 () const
{
    QString linestring = "";

    linestring.append ( getCodepointAsUtf8 () );
    linestring.append ( " ; " );

    for ( int i = 0; i < collations_.count (); i++ )
    {
        linestring.append ( collations_.at ( i ).toString () );
    }

    linestring.append ( " # " );
    linestring.append ( comment_ );
    return linestring;
}

/**
 * \brief Gets the version attribute of this letter.
 *
 * \return Version attribute of this letter.
 */
QString Letter::getVersionInfo () const
{
    return versionInfo_;
}

/**
 * \brief Compares letters by their codepoints.
 *
 * \param other The other letter.
 *
 * \return If true, the codepoints of this and other letter are identical in
 *         values, order and count.
 */
bool Letter::operator== ( const Letter& other ) const
{
    bool result = false;
    if ( other.getCodepoint () == codepoint_ )
    {
        result = true;
    }
    return result;
}

/**
 * \brief Compares letters by the first value of their codepoints.
 *
 * This method is not very accurate and is mainly implemented for the use by
 * QHash. If one of the letters codepoints is empty the result is always false.
 *
 * \param other The other letter.
 * \return True, if the first value of this letters codepoint is smaller than
 *         the first value of the other letters codepoint.
 */
bool Letter::operator< ( const Letter& other ) const
{
    bool result = false;
    if ( !other.getCodepoint ().isEmpty () && !codepoint_.isEmpty ()
            && codepoint_.at ( 0 ) < other.getCodepoint ().at ( 0 ) )
    {
        result = true;
    }
    return result;
}

/**
 * \brief Sets the block this letter belongs to.
 *
 * \param block The blocks name.
 */
void Letter::setBlock ( QString block )
{
    block_ = block;
}

/**
 * \brief Sets the codepoint.
 *
 * \param codepoint Codepoint.
 */
void Letter::setCodepoint ( QList<uint> codepoint )
{
    QString itemText ( "" );
    for ( int i = 0; i < codepoint.size (); ++i )
    {
        itemText.append( UCTools::unicodeToString(codepoint.at ( i )));
    }
    setText ( itemText );
    this->codepoint_ = codepoint;
}

/**
 * \brief Sets the collations.
 *
 * \param collations Collations.
 */
void Letter::setCollations ( QList<Collation> collations )
{
    this->collations_ = collations;
}

/**
 * \brief Sets the comment.
 *
 * \param comment Comment.
 */
void Letter::setComment ( QString comment )
{
    comment_ = comment;
    setToolTip ( comment );
}

/**
 * \brief Sets the version attribute of this letter.
 *
 * \param info Version.
 */
void Letter::setVersionInfo ( QString info )
{
    versionInfo_ = info;
}

/**
 * \brief Returns the background color for the letter in a UCQTableView.
 * If the letter has no collation, its background will be red.
 * If the letter has a collation, but no comment, its background will be yellow.
 * Letters with both, collation and comment, will have a white (no) background.
 * Letters with multiple codepoints, i.e. combined letters, will have a
 * green background.
 *
 * \return The background color for the Letter in the TableView.
 */
QBrush Letter::background ()
{
    QBrush brush ( Qt::white );

    if ( collations_.isEmpty () || collations_.at ( 0 ).isEmpty () )
        brush.setColor ( QColor ( 255, 192, 192, 255 ) );

    if ( comment_.isEmpty () )
        brush.setColor ( QColor ( 255, 255, 192, 255 ) );

    if ( codepoint_.size () > 1 )
        brush.setColor ( QColor ( 192, 255, 192, 255 ) );

    return brush;
}

///**
// * \brief Checks if the displayed letter is included in the given font.
// *
// * THIS METHOD DOES NOT WORK LIKE INTENDED.
// * Qt does not support checking for chars in fonts without using font merging
// * in the current version... According to the Qt bug logs this might be
// * available in Version 4.7.
// *
// * \param font The font to check for this letter.
// * \return If true is returned the letter is in the font included.
// */
//bool Letter::inFont( const QFont& font){
//    QString t = text();
//    if(t.isEmpty()){
//        return false;
//    }
//
//    QFont f(font);
//
//    bool result = true;
//    f.setStyleStrategy( QFont::NoFontMerging );
//    QFontMetrics fm(f);
//    QChar ch;
//    for(int i = 0; i < t.size(); ++i){
//        ch = t.at(i);
//        if(!fm.inFont(ch)){
//            result = false;
//        }
//    }
//    return result;
//}
