/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *     																	   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "TableIO.h"

/**
 * \brief Read tables with Unicode blocks.
 *
 * Lines must look like this: "DB80..DBFF; High Private Use Surrogates".
 * The first two hex values are interpreted as the block boundaries and the
 * rest of the line is used as the block name.
 *
 * \param filepath Path to file.
 * \param model TableModel object which takes the read blocks.
 * \param warnings If true, warnings are added when overwriting an existing
 *                 block.
 * \return If true, read was successful.
 */
bool TableIO::readBlockTable ( QString filepath, TableModel& model,
                               bool warnings )
{
    QFile file ( filepath );
    bool readOk = false;
    if ( !file.open ( QIODevice::ReadOnly | QIODevice::Text ) )
    {
        return readOk;
    }

    QString line;
    QTextStream in ( &file );
    QApplication::setOverrideCursor ( Qt::WaitCursor );

    int testedInt[ 2 ];
    bool testVar = false;
    while ( !in.atEnd () )
    {
        line = in.readLine ();

        //cut off any comments
        int idx = line.indexOf ( '#', 0, Qt::CaseInsensitive );
        if ( idx > -1 )
        {
            line.remove ( idx, line.size () - 1 );
        }

        // if line is not formatted right, skip it
        if ( line.isEmpty () || !line.at ( 0 ).isLetterOrNumber () )
        {
            continue;
        }

        // first split: 0 = block boundaries, 1 = name
        QStringList list = line.split ( ";" );
        if ( !list.isEmpty () )
        {
            // split the boundaries, results in three parts (0000..0000)
            // 0 = start, 1 = empty, 2 = end                  ^1  ^2 ^3
            QStringList startAndEnd = list[ 0 ].split ( "." );
            if ( list.size () > 1 && startAndEnd.size () > 2 )
            {
                // cut the rest of the name off, if it is unusually large
                if ( list.at ( 1 ).size () > MAXLENGTH )
                {
                    list[ 1 ].remove ( MAXLENGTH,
                                       list.at ( 1 ).size () - MAXLENGTH );
                    list[ 1 ].append ( "..." );
                }
                // make sure the conversion from string to unit is correct
                testedInt[ 0 ] = startAndEnd.at ( 0 ).toUInt ( &testVar, 16 );
                testedInt[ 1 ] = startAndEnd.at ( 2 ).toUInt ( &testVar, 16 );

                // add the block to the model
                if ( testVar )
                {
                    model.addBlock ( list.at ( 1 ).trimmed (),
                                     startAndEnd.at ( 0 ).toUInt ( 0, 16 ),
                                     startAndEnd.at ( 2 ).toUInt ( 0, 16 ),
                                     warnings );
                }
            }
        }
    }
    file.close ();

    QApplication::restoreOverrideCursor ();
    if ( file.error () == QFile::NoError )
    {
        readOk = true;
    }
    return readOk;
}

/**
 * \brief Reads a Unicode-table from a file and puts its data into the model.
 *
 * This method reads a Unicode-table from a plain text-file and extracts the
 * codepoint, the collation and the comment for every defined character.
 * The data is stored in Letter objects, which are managed in the TableModel.
 * The lines of the file to read should look like this:<br>
 * 0042  ; [.1225.0020.0008.0042] # LATIN CAPITAL LETTER B<br>
 * <codepoint> [<codepoint> ...] ; <collation> [<collation> ...] # comment <br>
 *
 * The table may have a version information like: <br>
 * \@version 5.2.0
 *
 * \param filepath Path to file to read.
 * \param model TableModel object that will contain the read letters.
 * \param overwrite If true, duplicates in the model are replaced.
 * \return If true, read was successful.
 *
 * \see Letter::setCodepoint()
 * \see Letter::addCollation()
 * \see Letter::setComment()
 * \see TableModel::addLetterList()
 */
bool TableIO::readTable ( QString filepath, TableModel& model, bool overwrite )
{
    QFile file ( filepath );
    bool readOk = false;
    if ( !file.open ( QIODevice::ReadOnly | QIODevice::Text ) )
    {
        return readOk;
    }

    QString line;
    QTextStream in ( &file );
    QList<Letter> letterList;

    int  testedInt;
    bool testVar;
    uint testedUInt;

    // collation regex
    QRegExp rx ( "([\\*]{0,}[0-9a-fA-F]{1,}.[0-9a-fA-F]{1,}."
                 "[0-9a-fA-F]{1,}.[0-9a-fA-F]{1,})" );

    QString version = " ";
    while ( !in.atEnd () )
    {
        line = in.readLine ();
        if ( line.isEmpty () )
        {
            continue;
        }

        // version information found
        if ( line.startsWith ( "@version" ) )
        {
            line.remove ( "@version" );
            version = line.trimmed ();
            if ( version.size () > MAXLENGTH )
            { // prevent abuse
                version.remove ( MAXLENGTH, version.size () - MAXLENGTH );
            }
            continue;
        }

        // no valid line, skip it
        if ( !line.at ( 0 ).isLetterOrNumber () )
        {
            continue;
        }

        // split into data and comment
        QStringList splittedLine = line.split ( "#" );
        QStringList codeAndCollation;
        if ( !splittedLine.isEmpty () )
        {
            codeAndCollation = splittedLine.at ( 0 ).split (
                                        ";", QString::SkipEmptyParts );
        }

        // no valid line, skip it
        if ( codeAndCollation.isEmpty () )
        {
            continue;
        }

        Letter letter;
        letter.setVersionInfo ( version );

        // split first part of data into codepoints
        QStringList codepoints =
                codeAndCollation.at ( 0 ).split ( " ", QString::SkipEmptyParts );

        // no codepoint == no valid letter
        if ( codepoints.isEmpty () )
        {
            continue;
        }

        // split codepoints, safely convert them to numbers and add them to the letter
        QList<uint> codepoint;
        for ( int i = 0; i < codepoints.size (); ++i )
        {
            testVar = false;
            testedUInt = codepoints.value ( i ).toUInt ( &testVar, 16 );
            if ( testVar )
                codepoint.append ( testedUInt );
        }
        letter.setCodepoint ( codepoint );

        // collations
        if ( codeAndCollation.size () > 1 )
        {
            QStringList collations;
            int pos = 0;

            // find all collations
            while ( ( pos = rx.indexIn ( codeAndCollation.value ( 1 ), pos ) )
                    != -1 )
            {
                collations << rx.cap ( 1 );
                pos += rx.matchedLength ();
            }

            for ( int i = 0; i < collations.count (); i++ )
            {
                Collation col;

                // split collation at dot into numbers
                QStringList collationlist =
                        collations.value ( i )
                            .split ( ".", QString::SkipEmptyParts );

                if ( !collationlist.isEmpty () )
                {
                    // check if the collation is variable and remove the *
                    if ( collationlist.at ( 0 ).at ( 0 ) == '*' )
                    {
                        col.setVariable ( true );
                        collationlist[ 0 ].remove ( 0, 1 );
                    }

                    // add weights/levels to the Collation object
                    for ( int j = 0; j < collationlist.size (); j++ )
                    {
                        testVar = false;
                        testedInt =
                                collationlist.at ( j ).toInt ( &testVar, 16 );
                        if ( testVar )
                        {
                            col.addWeight ( testedInt );
                        }
                    }
                }

                // add Collation to the Letter
                letter.addCollation ( col );
            }
        }

        // if a comment is available, add it to the Letter
        if ( splittedLine.size () > 1 )
        {
            QString comm = splittedLine.at ( 1 );

            // cut huge comments
            if ( comm.size () > MAXLENGTH )
            {
                comm.remove ( MAXLENGTH, comm.size () - MAXLENGTH );
                comm.append ( "..." );
            }
            letter.setComment ( comm.trimmed () );
        }

        // collect Letters in a list to add them to the model in one call
        letterList.append ( letter );
    }

    file.close ();

    if ( file.error () == QFile::NoError )
    {
        readOk = true;

        // add Letters to the model
        model.addLetterList ( letterList, overwrite );
    }
    return readOk;
}

/**
 * \brief Writes the items from visibleTable_ of the TableModel to a file.
 *
 * \param filepath Path to file to write.
 * \param model TableModel object whos visibleItems_ are written.
 * \return If true, write was successful.
 *
 * \see TableModel::getVisibleItems()
 * \see Letter::getLinestring()
 */
bool TableIO::writeVisibleTable ( QString filepath, TableModel& model )
{
    QFile file ( filepath );
    bool writeOk = false;

    if ( !file.open ( QIODevice::WriteOnly ) )
    {
        return writeOk;
    }

    QTextStream out ( &file );
    QApplication::setOverrideCursor ( Qt::WaitCursor );

    // get Letters to write
    QList<Letter*> list = model.getVisibleItems ();

    // write the Letters
    for ( int i = 0; i < list.size (); ++i )
    {
        out << list.at ( i )->getLinestring () << endl;
    }

    file.flush ();
    file.close ();

    QApplication::restoreOverrideCursor ();
    if ( file.error () == QFile::NoError )
    {
        writeOk = true;
    }
    return writeOk;
}
